/***************************************************************/
/*							       */
/*			     POLKAANIM.C		       */
/*							       */
/***************************************************************/
/*       Copyright 1993 Georgia Institute of Technology 
                     -- John T. Stasko                         */

#include "polka_local.H"

#include <X11/StringDefs.h>

/******************************************************************/
/* TANGO_anim_setup -- Create Pixmaps to hold animation frame(s). */
/*		       Whenever the frame dimensions are changed  */
/*		       (eg. by resizing the window) call this     */
/*		       function to create new Pixmaps and update  */
/*		       global TANGO__data information.		  */
/*								  */
/******************************************************************/
void
View::prepareAnim(Window wind)
{
   int	     old;
   Arg	     wargs[3];
   Dimension w, h;
   
   if (debug)
      printf("DEBUG: View::prepareAnim\n");

   destwin = wind;

   /* Create offscreen frame(s)...Pixmap(s) */

   if (window) XFreePixmap(_display, window);	/* Resizing old frame? */

   XtSetArg(wargs[0], XtNwidth, &w);
   XtSetArg(wargs[1], XtNheight, &h);
   XtGetValues(easel, wargs, 2);

   pixX = w;	/* Update global data */
   pixY = h;

   if ((created) &&                      // don't do this at start-up
       (resizeMode == ConstantAspect)) {
      rx = lx + (pixX / aspectx);
      ty = by + (pixY / aspecty);
      dx = rx-lx; 
      dy = ty-by; 
      }

   window = XCreatePixmap(_display, DefaultRootWindow(_display), w, h,
  	         	  DefaultDepth(_display, DefaultScreen(_display)));
   /* Pixmaps always start in an unknown state...clear it */

   old = set_color(WHITE);
   XFillRectangle(_display, window, inq_gc(), 0, 0, w, h);
   set_color(old);
}



/*******************************************************************/
/* TANGO_anim_next_frame -- Display next animation frame by	   */
/*			    PixBlting offscreen Pixmap to onscreen */
/*			    window.				   */
/*								   */
/*******************************************************************/
void
View::animNextFrame()
{
   if (debug)
      printf("DEBUG: View::animNextFrame\n");

   XCopyArea(_display, window, destwin, inq_gc(), 0, 0, pixX, pixY, 0, 0);
}

