
#include <stdio.h>
#include <stdlib.h>

#include "config.h"
#include <string.h>
#include "io.h"
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include "unix_defs.h"
#ifdef HAVE_WINDOWS_H
#include <windows.h>
#define sleep(x) Sleep(1000*x)
#else
#include <signal.h>
#include <netinet/in.h>
#endif
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif

#include "io_interface.h"
#include "io_internal.h"
extern int (*establish_server_connection_ptr)ARGS((IOFile iofile, int do_fallback));

static int
serverAtomicWrite(fd, buffer, length)
void* fd;
void *buffer;
int length;
{
     char *junk_result_str;
     int junk_errno;
     return os_server_write_func(fd, buffer, length, &junk_errno,
				&junk_result_str);
}

static void
usage_exit()
{
    printf("Usage:  format_cmd {checkpoint,restart,stats,ping}\n");
    exit(1);
}

static char *format_server_host = "\"not set\"";

static void
die_with_error(sig)
int sig;
{
    fprintf(stderr, "Format server %s not responding.  Timeout.\n", 
	    format_server_host);
    exit(1);
}

int
main(argc, argv)
int argc;
char **argv;
{
    IOContext context = create_IOcontext(NULL);
    IOFile iofile = (IOFile) context;
    char format_command_char = 'c';
    char format_dump_char = 'D';
    char format_read_char = 'R';
    char format_stats_char = 'S';
    char format_ping_char = 'p';
    char format_action_char;

    if (os_sockets_init_func != NULL) os_sockets_init_func();

    if (argc == 2) {
	if (strncmp(argv[1], "checkpoint", strlen(argv[1])) == 0) {
	    format_action_char = format_dump_char;
	} else if (strncmp(argv[1], "restart", strlen(argv[1])) == 0) {
	    format_action_char = format_read_char;
	} else if (strncmp(argv[1], "stats", strlen(argv[1])) == 0) {
	    format_action_char = format_stats_char;
	} else if (strncmp(argv[1], "ping", strlen(argv[1])) == 0) {
	    format_action_char = format_ping_char;
#ifndef HAVE_WINDOWS_H
	    signal(SIGALRM, die_with_error);
	    alarm(10);
#endif
	} else {
	    usage_exit();
	}
    } else {
	usage_exit();
    }

    /* 
     *  this mirrors code in server_acts.c, reproduced here to avoid
     *  exporting information
     */
    format_server_host = getenv("FORMAT_SERVER_HOST");
    if (format_server_host == NULL) {
	format_server_host = FORMAT_SERVER_HOST;	/* from configure */
    }

    if (establish_server_connection_ptr(iofile, 0) == 0) {
	printf("Failed to contact format server on host %s\n",
	       format_server_host);
	return 0;
    }

    if (serverAtomicWrite(iofile->server_fd, &format_command_char, 1) != 1) {
	fprintf(stderr, "command write failed\n");
	exit(1);
    }

    if (serverAtomicWrite(iofile->server_fd, &format_action_char, 1) != 1) {
	fprintf(stderr, "action write failed\n");
	exit(1);
    }

    if (format_action_char == 'S') {
	/* get stats */
	int len;
	char *buffer;
	if (serverAtomicRead(iofile->server_fd, &len, 4) != 4) {
	    fprintf(stderr, "length read failed\n");
	    exit(1);
	}
	len = ntohl(len);
	buffer = malloc(len);
	if (serverAtomicRead(iofile->server_fd, buffer, len) != len) {
	    fprintf(stderr, "stats block read failed\n");
	}
	dump_encoded_as_XML(context, buffer);
    } else if (format_action_char == 'p') {
	/* ping */
	char response;
	if (serverAtomicRead(iofile->server_fd, &response, 1) != 1) {
	    fprintf(stderr, "Ping response read failed.\n");
	    exit(1);
	}
	if (response != 'p') {
	    printf("Format server %s failed ping response.\n",
		   format_server_host);
	    exit(1);
	}
	printf("Format server %s is responsive.\n", format_server_host);
    }
    return 0;
}	
